#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 13:23:54 2020

@author: hasan
"""
# from pystac import (Catalog)
import argparse
from tqdm import trange,tqdm
import geopandas as gp
from shapely.geometry import Polygon
from rasterio.coords import BoundingBox
from rasterio import windows
from rasterio import warp
from rasterio.transform import from_bounds
from PIL import Image,ImageDraw
from skimage.morphology import dilation, square
from skimage.segmentation import watershed

from simplification.cutil import simplify_coords_vwp
from imantics import Mask
import numpy as np
import rasterio
from matplotlib import pyplot as plt
from rasterio.windows import Window
from typing import List, Tuple



def generate_polygon(bbox: List[float]) -> List[List[float]]:
    """
    Generates a list of coordinates forming a polygon.

    Parameters
    ----------
    bbox : List[float]
        A list representing the bounding box coordinates [xmin, ymin, xmax, ymax].

    Returns
    -------
    List[List[float]]
        A list of coordinates representing the polygon.
    """

    return [
        [bbox[0], bbox[1]],
        [bbox[2], bbox[1]],
        [bbox[2], bbox[3]],
        [bbox[0], bbox[3]],
        [bbox[0], bbox[1]]
    ]


def shape_polys(polyg):
    """Shapes the building polygons as a list of polygon lists.

    Parameters
    ----------
    polyg : List[Polygon]
        List of building polygons in numpy arrays.

    Returns
    -------
    List[List[Tuple(float, float)]]
        List of shaped polygons.
    """

    all_polys = []
    for poly in polyg:
        if len(poly) >= 3:
            f = poly.reshape(-1, 2)
            simplified_vw = simplify_coords_vwp(f, .3)
            if len(simplified_vw) > 2:
                mpoly = []
                # Rebuilding the polygon in the way that PIL expects the values [(x1,y1),(x2,y2)]
                for i in simplified_vw:
                    mpoly.append((i[0], i[1]))
                    # Adding the first point to the last to close the polygon
                mpoly.append((simplified_vw[0][0], simplified_vw[0][1]))
                all_polys.append(mpoly)
    return all_polys


def pol_to_np(pol: List[List[float]]) -> np.ndarray:
    """Converts a list of coordinates to a NumPy array.

    Parameters
    ----------
    pol : List[List[float]]
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    np.ndarray
        NumPy array of coordinates.
    """
    
    return np.array([list(l) for l in pol])


def pol_to_bounding_box(pol: List[List[float]]) -> BoundingBox:
    """Converts a list of coordinates to a bounding box.

    Parameters
    ----------
    pol : List[List[float]]
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    BoundingBox
        Bounding box of the coordinates.
    """

    arr = pol_to_np(pol)
    return BoundingBox(np.min(arr[:, 0]),
                       np.min(arr[:, 1]),
                       np.max(arr[:, 0]),
                       np.max(arr[:, 1]))
 

def reverse_coordinates(pol: List) -> List:
    """
    Reverse the coordinates in a polygon.

    Parameters
    ----------
    pol : list of list
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    list of list
        Reversed coordinates: [[y1, x1], [y2, x2], ..., [yN, xN]].
    """

    return [list(f[-1::-1]) for f in pol]


def get_areas(cols: dict) -> List:
    """
    Get areas from a collection of columns.

    Parameters
    ----------
    cols : dict
        A dictionary representing columns.

    Returns
    -------
    list
        A list of tuples containing area information. Each tuple has the format (iid, id1, id2).
    """

    areas = []
    for iid in cols:
        items = [x for x in cols[iid].get_all_items()]
        for i, id in enumerate(items):
            if i % 2 == 0 and i + 1 < len(items):
                areas.append((iid, items[i].id, items[i + 1].id))
    return areas


def to_index(wind_: Window) -> List[List[int]]:
    """
    Generates a list of index coordinates (row, col) for a given Window.

    Parameters
    ----------
    wind_ : Window
        The rasterio Window object specifying the region of interest.

    Returns
    -------
    List[List[int]]
        A list of index coordinates representing the corners of the Window.
    """

    return [
        [wind_.row_off, wind_.col_off],
        [wind_.row_off, wind_.col_off + wind_.width],
        [wind_.row_off + wind_.height, wind_.col_off + wind_.width],
        [wind_.row_off + wind_.height, wind_.col_off],
        [wind_.row_off, wind_.col_off]
    ]


def create_separation(labels: np.ndarray) -> np.ndarray:
    """Creates a mask for building spacing of close buildings.

    Parameters
    ----------
    labels : np.ndarray
        Numpy array where each building's pixels are encoded with a certain value.

    Returns
    -------
    np.ndarray
        Mask for building spacing.
    """
    
    tmp = dilation(labels > 0, square(20))
    tmp2 = watershed(tmp, labels, mask=tmp, watershed_line=True) > 0
    tmp = tmp ^ tmp2
    tmp = dilation(tmp, square(5))
    msk1 = np.zeros_like(labels, dtype='bool')

    for y0 in range(labels.shape[0]):
        for x0 in range(labels.shape[1]):
            if not tmp[y0, x0]:
                continue
            if labels[y0, x0] == 0:
                sz = 15
            else:
                sz = 20
            uniq = np.unique(labels[max(0, y0 - sz):min(labels.shape[0], y0 + sz + 1),
                             max(0, x0 - sz):min(labels.shape[1], x0 + sz + 1)])
            if len(uniq[uniq > 0]) > 1:
                msk1[y0, x0] = True
    return msk1


def make_instance_mask(all_polys: List[List[Tuple[float, float]]], size: int) -> Image:
    """Encodes each building polygon with a value in the mask.

    Parameters
    ----------
    all_polys : List[List[Tuple[float, float]]]
        List of building polygons.
    size : int
        Width and height of the square mask.

    Returns
    -------
    Image
        Instance mask.
    """

    bg=np.zeros((size,size)).astype(np.uint8)
    bg=Image.fromarray(bg).convert('L')
    shift=255-len(all_polys)
    for i,poly in enumerate(all_polys):
        ImageDraw.Draw(bg).polygon(poly,outline=shift+i,fill=shift+i)
    return bg



from sklearn.model_selection import StratifiedKFold
import pandas as pd

class Splitting():
    """
    Splitting class for all split methods
 
    Args:
        df (pd.DataFrame): Input DataFrame.
        nfolds (int): Number of folds.
        seed (int): Random seed for reproducibility.
        id_key (str): Key representing the identifier for grouping.
        split_key (str): Key for stratification.
        label_keys (list): List of keys for labels.
        verbose (bool, default=False): If True, print fold statistics.
    """
    def __init__(self, df, nfolds, seed, id_key, split_key, label_keys, verbose=False):
        self.df = df
        self.nfolds = nfolds
        self.seed = seed
        self.id_key = id_key
        self.split_key = split_key
        self.label_keys = label_keys
        self.verbose = verbose
        
    def split_K_stratified_folds(self):
        """
        Split the DataFrame into K stratified folds based on specified keys.
        
        Returns:
            pd.DataFrame: DataFrame with an additional 'fold' column indicating the fold number.
        """
        X = self.df.groupby(self.id_key)[self.split_key].first().index.values
        y = self.df.groupby(self.id_key)[self.split_key].first().values
        skf = StratifiedKFold(n_splits=self.nfolds, random_state=self.seed, shuffle=True)
    
        for i, (tfold, vfold) in enumerate(skf.split(X, y)):
            self.df.loc[self.df[self.id_key].isin(X[vfold]), 'fold'] = int(i)
    
        folds = [int(fold) for fold in self.df.groupby('fold').first().index.values]
        if self.verbose:
            for fold in folds:
                for label_key in self.label_keys:
                    print(f'fold:\t{fold}')
                    print(f'Label Key:{label_key}')
                    print(self.df.loc[self.df['fold'] == fold].set_index(['fold', label_key]).groupby(level=label_key).count())
        self.df.reset_index(drop=True, inplace=True)
        return self.df
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  1 16:02:07 2021

@author: hasan
"""

import torch
import torch.nn as nn
import torch._utils
import torch.nn.functional as F


class BNReLU(nn.Sequential):
    def __init__(self,in_channels):
        super().__init__(nn.BatchNorm2d(in_channels),
                         nn.ReLU())

class SpatialGather_Module(nn.Module):
    """
        Aggregate the context features according to the initial 
        predicted probability distribution.
        Employ the soft-weighted method to aggregate the context.
    """
    def __init__(self, cls_num=0, scale=1):
        super(SpatialGather_Module, self).__init__()
        self.cls_num = cls_num
        #self.scale = scale

    def forward(self, feats, probs):
        batch_size, c, h, w = probs.size(0), probs.size(1), probs.size(2), probs.size(3)
        probs = probs.view(batch_size, c, -1)
        feats = feats.view(batch_size, feats.size(1), -1)
        feats = feats.permute(0, 2, 1) # batch x hw x c 
        probs = F.softmax(probs, dim=2) 

        ocr_context = torch.matmul(probs, feats).permute(0, 2, 1).unsqueeze(3)# batch x k x c
        return ocr_context


class ObjectAttentionBlock2D(nn.Module):
    '''
    The basic implementation for object context block
    Input:
        N X C X H X W
    Parameters:
        in_channels       : the dimension of the input feature map
        key_channels      : the dimension after the key/query transform
        scale             : choose the scale to downsample the input feature maps (save memory cost)
        bn_type           : specify the bn type
    Return:
        N X C X H X W
    '''
    def __init__(self, 
                 in_channels, 
                 key_channels,
                 cls_num = 1):
        super(ObjectAttentionBlock2D, self).__init__()
        self.in_channels = in_channels
        self.key_channels = key_channels
        self.cls_num = cls_num
        self.f_pixel = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.key_channels),
            nn.Conv2d(in_channels=self.key_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.key_channels),
        )
        
        self.f_object = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.key_channels),
            nn.Conv2d(in_channels=self.key_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.key_channels),
        )
        
        self.f_down = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.key_channels),
        )
        self.f_up = nn.Sequential(
            nn.Conv2d(in_channels=self.key_channels, out_channels=self.in_channels,
                kernel_size=1, stride=1, padding=0, bias=False),
            BNReLU(self.in_channels),
        )

    def forward(self, x, proxy):
        batch_size, h, w = x.size(0), x.size(2), x.size(3)

        query = self.f_pixel(x).view(batch_size, self.key_channels, -1)
        query = query.permute(0, 2, 1)
        key = self.f_object(proxy).view(batch_size, self.key_channels, -1)
        value = self.f_down(proxy).view(batch_size, self.key_channels, -1)
        value = value.permute(0, 2, 1)

        sim_map = torch.matmul(query, key)
        sim_map = (self.key_channels**-.5) * sim_map
        sim_map = F.softmax(sim_map, dim=-1) 
        if(self.cls_num ==1):
            sim_map = sim_map[...,0].unsqueeze(2)
            value = value[:,0,...].unsqueeze(1)
        print(sim_map.min())
        print(sim_map.max())
        print(sim_map.size())
        # add bg context ...
        context = torch.matmul(sim_map, value)
        context = context.permute(0, 2, 1).contiguous()
        context = context.view(batch_size, self.key_channels, *x.size()[2:])
        context = self.f_up(context)

        return context


class SpatialOCR_Module(nn.Module):
    """
    Implementation of the OCR module:
    We aggregate the global object representation to update the representation for each pixel.
    """
    def __init__(self, 
                 in_channels, 
                 key_channels, 
                 out_channels, 
                 dropout=0.1,
                 cls_num = 1):
        super(SpatialOCR_Module, self).__init__()
        self.object_context_block = ObjectAttentionBlock2D(in_channels, 
                                                           key_channels,
                                                           cls_num)
        _in_channels = 2 * in_channels

        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(_in_channels, out_channels, kernel_size=1, padding=0, bias=False),
            BNReLU(out_channels),
            nn.Dropout2d(dropout)
        )

    def forward(self, feats, proxy_feats):
        context = self.object_context_block(feats, proxy_feats)
        output = self.conv_bn_dropout(torch.cat([context, feats], 1))

        return output

class OCR(nn.Module):
    def __init__(self,
                 in_channels,
                 key_channels=256,
                 out_channels=512,
                 dropout = 0 ,
                 num_cls = 1):
        super().__init__()
        
        
        self.aux_head = nn.Sequential(nn.Conv2d(in_channels = in_channels,
                                                out_channels = in_channels,
                                                kernel_size = 3,
                                                padding = 1),
                                      nn.BatchNorm2d(in_channels),
                                      nn.ReLU(inplace = True),
                                      nn.Conv2d(in_channels = in_channels,
                                                out_channels = num_cls if(num_cls!=1) else 2,
                                                kernel_size = 1),
                                      )
        self.conv3x3_ocr = nn.Sequential(nn.Conv2d(in_channels = in_channels,
                                                   out_channels = out_channels,
                                                   kernel_size = 3,
                                                   padding = 1),
                                         nn.BatchNorm2d(out_channels),
                                         nn.ReLU(inplace = True))
        self.ocr_gather_head = SpatialGather_Module(num_cls)
        
        self.ocr_distri_head = SpatialOCR_Module(out_channels,
                                                 key_channels,
                                                 out_channels,
                                                 dropout,
                                                 num_cls)
    def forward(self,feats):
        
        probs = self.aux_head(feats)
        feats = self.conv3x3_ocr(feats)
        context = self.ocr_gather_head(feats,probs)
        aug_feats = self.ocr_distri_head(feats,context)
        return aug_feats
        
'''
t = torch.rand((1,128,16,16)).cuda()
m = OCR(128)
m.eval()
m.cuda()
y = m(t)
print(y.size())
'''


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 17:56:35 2021

@author: hasan
"""
#useful functions :D
import geopandas as gpd
import rasterio as rio
from shapely.geometry import Polygon,box
from mercantile import bounds
from supermercado.burntiles import burn
from tqdm import tqdm,trange
from math import ceil
from pystac import (Catalog)
import argparse
from tqdm import trange,tqdm
import geopandas as gp
from shapely.geometry import Polygon
from rasterio.coords import BoundingBox
from rasterio import windows
from rasterio import warp
from rasterio.transform import from_bounds
from PIL import Image,ImageDraw
from skimage.morphology import dilation, square, watershed
import solaris as sol
from simplification.cutil import simplify_coords_vwp
from imantics import Mask
import numpy as np
import rasterio
from matplotlib import pyplot as plt
from rasterio.windows import Window
from typing import List, Tuple

global our_crs
our_crs = 'WGS84'

def get_fitting_box_from_box(bbox,width,height):
    minx,miny,maxx,maxy = bbox.bounds
    cx = (minx+maxx) // 2
    cy = (miny+maxy) // 2
    
    box_w = maxx - minx 
    box_h = maxy - miny
    
    fbox_w = ceil(box_w / 512) * 512
    fbox_h = ceil(box_h / 512) * 512

    gap_left = fbox_w // 2 - (cx - max(0,cx - fbox_w // 2))
    gap_right = fbox_w // 2 - (min(width,cx + fbox_w // 2) - cx)
    gap_up = fbox_h // 2 - (cy - max(0,cy - fbox_h // 2))
    gap_down = fbox_h // 2 - (min(height,cy + fbox_h // 2) - cy)

    fb_minx = cx - (fbox_w // 2 + gap_right - gap_left) 
    fb_maxx = cx + (fbox_w // 2 + gap_left - gap_right)
    fb_miny = cy - (fbox_h // 2 + gap_down - gap_up)
    fb_maxy = cy + (fbox_h // 2 + gap_up - gap_down)

    fbox = box(fb_minx,fb_miny,fb_maxx,fb_maxy)
    return fbox

def poly_lonlat2pix(poly,bbox_bounds,img,width = None,height = None):
  if(img is not None):
    h,w = img.shape[:2]
  elif(None not in set([width,height])):
    h,w = height,width
  else:
    raise ValueError('Either Image or height and width should not be None')
  transform = rio.transform.from_bounds(*bbox_bounds,w,h)
  xs,ys = poly.exterior.xy
  rows,cols = rio.transform.rowcol(transform,xs,ys)
  coords = list(zip(cols,rows))
  return Polygon(coords)

def get_tiles_xyz(gdf,zoom_level):
  gdf_geo = gdf.__geo_interface__
  features  = gdf_geo['features']
  tiles_xyz = burn(features,zoom_level)
  return tiles_xyz

def get_tiles_xyz_fast(gdf,zoom_level):
    part =100000
    l = len(gdf)
    all_tiles = set()
    if(l<part):
        part = l 
    for i in trange(0,l,part):
        c = min(part,l-i)
        tiles_xyz = get_tiles_xyz(gdf[i:i+c],zoom_level)
        tiles_xyz = list(map(tuple,tiles_xyz))
        all_tiles.update(tiles_xyz)
    return list(all_tiles)

def get_covering_tiles(gdf,zoom_level):
  tiles_xyz = get_tiles_xyz_fast(gdf,zoom_level)
  tiles_bboxs = []
  for xyz in tiles_xyz:
    _b = bounds(xyz)
    tiles_bboxs.append(box(*_b))
  return tiles_xyz,tiles_bboxs

def poly2gdf(poly,crs):
  return gpd.GeoDataFrame({'geometry':[poly]},crs=our_crs)

def encode_tile_id(tile_xyz):
  return 't_{}_{}_{}'.format(*tile_xyz)

def encode_tiles_ids(tiles_xyz):
  ids = [encode_tile_id(tile_xyz) for tile_xyz in tiles_xyz]
  return ids

def encode_multi_id(ids):
  return '&'.join(ids)

def decode_id(tile_id):
  return [*map(int,tile_id.lstrip('t_').split('_'))]

def decode_multi_id(multi_tile_id):
  return [decode_id(tid) for tid in multi_tile_id.split('&')]

#very specific functions
def _compress(gdf,_keys,id_key = 'FM_RE_ID'):
  keys = gdf.keys()
  _keys.extend(['tile_id'])
  assert(False not in [(k in keys ) for k in _keys]),'Missing keys from {}'.format(_keys)
  _keys[-1] = 'covering_tiles_ids'
  #print(_keys)
  dic = {k : [] for k in _keys}
  uniq_fm_ids = gdf[id_key].unique()
  for i,fm_id in enumerate(uniq_fm_ids):
    query = gdf.loc[gdf[id_key] == fm_id] 
    values = [query[_k][query.index[0]] if(_k != id_key) else fm_id for _k in _keys[:-1] ]
    values.append(encode_multi_id(list(query['tile_id'])))
    for k,v in [*zip(_keys,values)]:dic[k].append(v)
  return gpd.GeoDataFrame(dic,crs=gdf.crs)

def get_covering_tiles_perloc(gdf,zoom_level,id_key = 'FM_RE_ID'):
  crs = gdf.crs
  tiles_xyz,tiles_bboxs = get_covering_tiles(gdf,zoom_level)
  tiles_ids = encode_tiles_ids(tiles_xyz)
  tiles_gdf = gpd.GeoDataFrame({'tile_id' : tiles_ids,
                                'geometry': tiles_bboxs,
                                },
                               crs = crs)
  joined = gpd.sjoin(gdf,tiles_gdf, how='left', op='intersects')
  del joined['index_right']
  joined.reset_index(inplace = True,drop = True) 
  return _compress(joined,[*gdf.keys()],id_key),tiles_gdf

def get_sample(fm_id,gdf,col = 'FMID_RE_ID'):
  query = gdf.loc[gdf[col]==fm_id]
  return query.loc[query.index[0]]

def get_tiles_bboxs(tiles_ids,gdf,col = 'tile_id'):
  tiles_bboxs=[]
  for tile_id in tiles_ids:
    query = gdf.loc[gdf[col] == tile_id]
    tiles_bboxs.append(query['geometry'][query.index[0]])
  return tiles_bboxs



def get_areas(cols: dict) -> List:
    """
    Get areas from a collection of columns.

    Parameters
    ----------
    cols : dict
        A dictionary representing columns.

    Returns
    -------
    list
        A list of tuples containing area information. Each tuple has the format (iid, id1, id2).
    """

    areas = []
    for iid in cols:
        items = [x for x in cols[iid].get_all_items()]
        for i, id in enumerate(items):
            if i % 2 == 0 and i + 1 < len(items):
                areas.append((iid, items[i].id, items[i + 1].id))
    return areas


def to_index(wind_: Window) -> List[List[int]]:
    """
    Generates a list of index coordinates (row, col) for a given Window.

    Parameters
    ----------
    wind_ : Window
        The rasterio Window object specifying the region of interest.

    Returns
    -------
    List[List[int]]
        A list of index coordinates representing the corners of the Window.
    """

    return [
        [wind_.row_off, wind_.col_off],
        [wind_.row_off, wind_.col_off + wind_.width],
        [wind_.row_off + wind_.height, wind_.col_off + wind_.width],
        [wind_.row_off + wind_.height, wind_.col_off],
        [wind_.row_off, wind_.col_off]
    ]


def get_winNmask(rst: rasterio.io.DatasetReader, x: int, y: int, width: int, height: int, polys: List[Polygon]) -> Tuple[np.ndarray, np.ndarray]:
    """This function crops a chip for a specified width and height and generates the mask of buildings with borders. It requires solaris framework

    Parameters
    ----------
    rst : rasterio.io.DatasetReader
        Raster which is the tif image to be cropped open using rasterio.open(tif url)
    x : int 
        X pixel coordinate of the top-left corner of the window.
    y : int 
        Y pixel coordinate of the top-left corner of the window.
    width : int 
        The width of the window to be chipped
    height : int 
        The height of the window to be chipped
    polys : List[Polygon] 
        the polygons of the whole raster image, in opencities ai the data was stored cleanly in a pystac and i was able
        to insert the polygons of the buildings in geo-pandas dataframe , the polys parameter was called by geo-pandas-df-polys.geometry
        so this is very specific and i didnt test other formats

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        aoo : np.array of RGBA channels of the image chipped(cropped)
        fbc_mask : an np array of shape [width,height,2] it contains the mask of buildings and borders
    """

    win = Window(x,y,width,height)
    bbox = windows.bounds(win,rst.transform)
    pol = generate_polygon(bbox)
    pol_np = np.array(pol)
    coords_transformed = warp.transform(rst.crs,{'init': 'epsg:4326'},pol_np[:,0],pol_np[:,1])
    
    ct=coords_transformed
    l,b,r,t=min(ct[0]),min(ct[1]),max(ct[0]),max(ct[1])
    coords_transformed = [[r,c] for r,c in zip(coords_transformed[0],coords_transformed[1])]
    tfm1=from_bounds(l,b,r,t,height,width)
    coords_transformed= Polygon(coords_transformed)
    arr_win = rst.read(window=win)
    
    all_polys = [poly for poly in polys if poly.intersects(coords_transformed)]
    all_polys_gdf = gp.GeoDataFrame(geometry=all_polys,crs='epsg:4326')
    fbc_mask = sol.vector.mask.df_to_px_mask(df=all_polys_gdf,
                                            channels=['footprint', 'boundary'],
                                            affine_obj=tfm1, shape=(width,height),
                                            boundary_width=4, boundary_type='inner',meters=False)
    aoo=np.array(arr_win)
    return aoo,fbc_mask


def create_separation(labels: np.ndarray) -> np.ndarray:
    """Creates a mask for building spacing of close buildings.

    Parameters
    ----------
    labels : np.ndarray
        Numpy array where each building's pixels are encoded with a certain value.

    Returns
    -------
    np.ndarray
        Mask for building spacing.
    """
    
    tmp = dilation(labels > 0, square(20))
    tmp2 = watershed(tmp, labels, mask=tmp, watershed_line=True) > 0
    tmp = tmp ^ tmp2
    tmp = dilation(tmp, square(5))
    msk1 = np.zeros_like(labels, dtype='bool')

    for y0 in range(labels.shape[0]):
        for x0 in range(labels.shape[1]):
            if not tmp[y0, x0]:
                continue
            if labels[y0, x0] == 0:
                sz = 15
            else:
                sz = 20
            uniq = np.unique(labels[max(0, y0 - sz):min(labels.shape[0], y0 + sz + 1),
                             max(0, x0 - sz):min(labels.shape[1], x0 + sz + 1)])
            if len(uniq[uniq > 0]) > 1:
                msk1[y0, x0] = True
    return msk1


def make_instance_mask(all_polys: List[List[Tuple[float, float]]], size: int) -> Image:
    """Encodes each building polygon with a value in the mask.

    Parameters
    ----------
    all_polys : List[List[Tuple[float, float]]]
        List of building polygons.
    size : int
        Width and height of the square mask.

    Returns
    -------
    Image
        Instance mask.
    """

    bg=np.zeros((size,size)).astype(np.uint8)
    bg=Image.fromarray(bg).convert('L')
    shift=255-len(all_polys)
    for i,poly in enumerate(all_polys):
        ImageDraw.Draw(bg).polygon(poly,outline=shift+i,fill=shift+i)
    return bg


def tile_area(area: str, iid: str, lid: str, tile_size: int, ipth: str, mpth: str, cols: Catalog) -> None:
    """Tiles a certain area (tif image) and generates the corresponding tile mask.
    Parameters
    ----------
    area : str
        Area ID (example: acc-665946).
    iid : str
        Image ID in the pystac.
    lid : str
        Label ID in the pystac.
    tile_size : int
        Size of width and height of the tile (use only square).
    ipth : str
        Directory to save image tiles.
    mpth : str
        Directory to save the masks.
    cols: Catalog
        Get all the columns in the catalog(main_children), it is how the pystac is organized
    """
    item=cols[area].get_item(id=iid)
    label=cols[area].get_item(id=lid)
    lbl_gfd=gp.read_file(label.make_asset_hrefs_absolute().assets['labels'].href)
    polys=lbl_gfd.geometry
    tif_url = item.assets['image'].href
    if tif_url.startswith("./"):
        tif_url = '/'.join(item.to_dict()['links'][1]['href'].split("/")[:-1])+tif_url[1:]
    rst = rasterio.open(tif_url)
    wid,hei=rst.width,rst.height
    for i in trange(0,tile_size*int(hei/tile_size),tile_size):
        for j in range(0,tile_size*int(wid/tile_size),tile_size):
                aoo,mask=get_winNmask(rst,j,i,tile_size,tile_size,polys)
                uniq=np.unique(aoo)
                if(np.all(uniq==0)):continue
                uniq=np.unique(mask[:,:,0])
                if(np.all(uniq==0)):
                    spacing=np.zeros((tile_size,tile_size)).astype(np.uint8)
                else:
                    getit=mask[:,:,0]^mask[:,:,1]
                    polyg = Mask(getit).polygons()
                    extracted=shape_polys(polyg)
                    labels=np.array(make_instance_mask(extracted,tile_size))
                    spacing=np.array(create_separation(labels)).astype(np.uint8)*255
                aoo=np.moveaxis(aoo,0,2)
                final_mask=spacing
                final_mask[(np.where(mask[:,:,0]>0))]=np.uint8(64)
                final_mask[(np.where(mask[:,:,1]>0))]=np.uint8(128)
                img=Image.fromarray(aoo).convert('RGB')
                fmsk=Image.fromarray(final_mask)
                img.save('/'.join([ipth,f'{area}_{iid}_{i}_{j}.png']))
                fmsk.save('/'.join([mpth,f'{area}_{iid}_{i}_{j}_mask.png']))


import torch
from torch.optim.lr_scheduler import (
    CosineAnnealingWarmRestarts, 
    CosineAnnealingLR, 
    MultiStepLR,
    LambdaLR,
    OneCycleLR,
    ConstantLR,
    ReduceLROnPlateau,
    LinearLR,
    ExponentialLR,
    MultiplicativeLR,
    StepLR,
    PolynomialLR,
    ChainedScheduler,
    SequentialLR,
    CyclicLR,
    LRScheduler
    )
# polylr
from .polylr import PolyLR,PolyLR_WWP

needs_total_iters = {
    'cosine-anneal' : 'T_max',
    'onecyclelr' : ' total_steps',
}

scheduler_mapping = {
    'polylr' : PolyLR,
    'polylrwwp' : PolyLR_WWP,
    'multi-steplr' : MultiStepLR,
    'cosine-anneal' : CosineAnnealingLR,
    'cosine-anneal-wr' : CosineAnnealingWarmRestarts,
    'lambdalr' : LambdaLR,
    'onecyclelr' : OneCycleLR,
    'constantlr' : ConstantLR,
    'linearlr' : LinearLR,
    'exponentiallr' : ExponentialLR,
    'reducelronplateau':ReduceLROnPlateau,
    'multiplicativelr':MultiplicativeLR,
    'steplr':StepLR,
    'polynomiallr':PolynomialLR,
    'chainedscheduler':ChainedScheduler,
    'sequentiallr':SequentialLR,
    'cycliclr':CyclicLR,
}


def get_scheduler(name :str, optimizer :torch.optim.Optimizer, **kwargs) -> LRScheduler:
    """
    This function returns the scheduler given its name

    @param name: name of the scheduler
    @param params: Optimeizer to schedule
    @param **kwargs:named parameters for the sheduler

    @type name:str
    @type optimizer: torch.optim.Optimizer
    
    @return: torch.optim.lr_scheduler.LRScheduler

    """
    name = name.lower()
    if name not in scheduler_mapping.keys():
        raise ValueError(f'scheduler {name} is not implemented!!!')
    
    return scheduler_mapping[name](optimizer=optimizer, **kwargs)

class AutoScheduler:
    """
    This class automate the step of the schechduler, the user will call scheduler.step on every iteration and the scheduler once required
    """
    def __init__(
            self,
            name :str,
            optimizer :torch.optim.Optimizer,
            data_loader = None,
            total_epochs :int = None,
            iters_per_epoch :int|None = None,
            mode = 'per_epoch', 
            **kwargs
            ):
        
        """
            This function returns the scheduler given its name

            @param name: name of the scheduler
            @param params: Optimeizer to schedule
            @param data_loader: (optional) used to find the iters_per_epoch
            @param total_epochs: (optional)
            @param iters_per_epoch: (optional)
            @param **kwargs:named parameters for the sheduler

            @type name:str
            @type optimizer: torch.optim.Optimizer


        """
        
        assert isinstance(total_epochs,int), "the total_epochs must be an int"
        if data_loader is not None:
            self.iters_per_epoch = len(data_loader)
        elif isinstance(iters_per_epoch,int):
            self.iters_per_epoch = iters_per_epoch
        else:
            raise ValueError(f'please correctly provide data_loader or iter_per_epoch')
        
        self.total_iters = total_epochs * self.iters_per_epoch 
        self.iter_counter = 0

        self.set_mode(mode)

        if name.startswith('polylr'):
            kwargs['epochs'] = total_epochs if mode == 'per_epoch' else self.total_iters
        if name in needs_total_iters.keys():
            kwargs[needs_total_iters[name]] = self.total_iters

        self.scheduler = get_scheduler(name=name,optimizer=optimizer,**kwargs)
    
    def set_mode(self,mode):
        """
        sets mode of the iter_counter
        @param mode: can be only 'per_epoch' or 'per_iter'
        """
        assert mode in ['per_epoch','per_iter']
        self.mode = mode
        self.set_stepsize()

    def set_stepsize(self):
        """define stepsize"""
        if self.mode == 'per_epoch':
            self.stepsize = self.iters_per_epoch
        else:
            self.stepsize = 1

    def step(self):
        """
            done in each iteration, steps only when iter_counter == self.stepsize 
        """
        self.iter_counter += 1

        if self.iter_counter > self.total_iters:
            return
        
        if self.iter_counter % self.stepsize == 0:
            self.scheduler.step()





import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from segmentation_models_pytorch.utils import functional as F_
#from lovasz import LovaszLoss,BinaryLovaszLoss

from segmentation_models_pytorch.utils.losses import JaccardLoss,DiceLoss


from segmentation_models_pytorch.losses import DiceLoss,SoftBCEWithLogitsLoss


def _neg_loss(pred, gt):
  ''' Modified focal loss. Exactly the same as CornerNet.
      Runs faster and costs a little bit more memory
    Arguments:
      pred (batch x c x h x w)
      gt_regr (batch x c x h x w)
  '''
  pos_inds = gt.eq(1).float()
  neg_inds = gt.lt(1).float()

  neg_weights = torch.pow(1 - gt, 4)

  loss = 0

  pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
  neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

  num_pos  = pos_inds.float().sum()
  pos_loss = pos_loss.sum()
  neg_loss = neg_loss.sum()

  if num_pos == 0:
    loss = loss - neg_loss
  else:
    loss = loss - (pos_loss + neg_loss) / num_pos
  return loss









def _soft_neg_loss(pred, gt,eps=0.2):
  ''' Modified focal loss to soft focal loss for noisy labels :)
    Arguments:
      pred (batch x c x h x w)
      gt_regr (batch x c x h x w)
  '''
  mask = gt.ge(0).float() * gt.le(1).float()

  pos_inds = gt.ge(1 - eps).float() * mask
  neg_inds = gt.lt(1 - eps).float() * mask

  neg_weights = torch.pow(1 - gt, 4)

  loss = 0

  pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
  neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

  num_pos  = pos_inds.float().sum()
  pos_loss = pos_loss.sum()
  neg_loss = neg_loss.sum()

  if num_pos == 0:
    loss = loss - neg_loss
  else:
    loss = loss - (pos_loss + neg_loss) / num_pos
  return loss










class KeypointFocalLoss(nn.Module):
  '''nn.Module warpper for focal loss'''
  __name__ = 'keypoint_focal_loss'
  def __init__(self):
    super(KeypointFocalLoss, self).__init__()
    self.neg_loss = _neg_loss

  def forward(self, out, target):
    out = out.clamp_max(0.999).clamp_min(0.0001)
    return self.neg_loss(out, target)







class SmoothKeypointFocalLoss(nn.Module):
  '''nn.Module warpper for smooth focal loss'''
  __name__ = 'keypoint_smooth_focal_loss'
  def __init__(self,eps):
    super(SmoothKeypointFocalLoss, self).__init__()
    self.eps = eps
    self.soft_neg_loss = _soft_neg_loss

  def forward(self, out, target):
    out = out.clamp_max(0.999).clamp_min(0.0001)
    return self.soft_neg_loss(out, target,self.eps)


class DiceBCELoss(nn.Module):

  def __init__(self,alpha_beta=(1.0,1.0),mode='binary',bce_pos_weight=1.0):
    super().__init__()
    
    self.dice = DiceLoss(mode=mode,from_logits=True,smooth=0,eps=1e-7)
    self.bce = SoftBCEWithLogitsLoss(pos_weight=bce_pos_weight)

    self.alpha ,self.beta = alpha_beta
  
  def forward(self,pred,gt):

    dice_loss = self.dice(pred,gt)
    bce_loss = self.bce(pred,gt)

    loss = self.alpha * dice_loss + self.beta * bce_loss

    return loss
 


class MaskedBCELoss(nn.Module):

  def __init__(self,device='cpu'):
    super().__init__()
    self.bce_criterion = nn.BCEWithLogitsLoss(reduction='none')
    self.device = device 
  
  def forward(self,pred,gt,mask):
    B,C,H,W = pred.shape
    loss = torch.zeros_like(pred, requires_grad=True).to(device=self.device)

    pred = torch.masked_select(pred,mask)
    gt = torch.masked_select(gt,mask)

    if pred.shape[0] != 0:
      bce_loss = self.bce_criterion(pred,gt)
      loss = loss.masked_scatter(mask,bce_loss)

    loss = loss.view(B,C,-1)
    mask = mask.view(B,C,-1)
    counts = mask.float().sum(dim=-1).clamp_min(1.0)

    loss = loss.sum(dim=-1) / counts
    loss = loss.mean()

    return loss















    



class Activation(nn.Module):

    def __init__(self, name, **params):

        super().__init__()

        if name is None or name == 'identity':
            self.activation = nn.Identity(**params)
        elif name == 'sigmoid':
            self.activation = nn.Sigmoid()
        elif name == 'softmax2d':
            self.activation = nn.Softmax(dim=1, **params)
        elif name == 'softmax':
            self.activation = nn.Softmax(**params)
        elif name == 'logsoftmax':
            self.activation = nn.LogSoftmax(**params)
        elif name == 'argmax':
            self.activation = ArgMax(**params)
        elif name == 'argmax2d':
            self.activation = ArgMax(dim=1, **params)
        elif callable(name):
            self.activation = name(**params)
        else:
            raise ValueError('Activation should be callable/sigmoid/softmax/logsoftmax/None; got {}'.format(name))

    def forward(self, x):
        return self.activation(x)
    



class FocalLoss(nn.Module):
    __name__='focal_loss'
    def __init__(self, alpha=1, gamma=2, logits=True, reduce=True):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.logits = logits
        self.reduce = reduce

    def forward(self, inputs, targets):
        if self.logits:
            BCE_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduce=False)
        else:
            BCE_loss = F.binary_cross_entropy(inputs, targets, reduce=False)
        pt = torch.exp(-BCE_loss)
        F_loss = self.alpha * (1-pt)**self.gamma * BCE_loss

        if self.reduce:
            return torch.mean(F_loss)
        else:
            return F_loss



class BCELoss(nn.Module):
    __name__='bce_loss'
    def __init__(self,logits=True):
        super().__init__()
        self.logits = logits
        self.bce = nn.BCELoss(reduction='mean') if(not(logits)) else nn.BCEWithLogitsLoss(reduction='mean')
    def forward(self,y_pr,y_gt):
        bce = self.bce(y_pr,y_gt)
        return bce


class CELoss(nn.Module):
    __name__='ce_loss'
    def __init__(self,weights=[1.,5.,5.,1.]):
        super().__init__()
        self.ce = nn.CrossEntropyLoss(reduction='mean',weight=weights)
    def forward(self,y_pr,y_gt):
        ce = self.ce(y_pr,y_gt)
        return ce
    



class DiceBCELoss(nn.Module):
    __name__='dice_bce_loss'
    def __init__(self,weights=[1.,1.],activation='sigmoid'):
        super().__init__()
        self.bce = BCELoss()
        self.dice = DiceLoss(activation=activation)
        self.w1,self.w2=weights[0],weights[1]
    def forward(self,y_pr,y_gt):
        return self.w2*self.bce(y_pr,y_gt) + self.w1*self.dice(y_pr, y_gt)
    



class JaccardBCELoss(nn.Module):
    __name__='jaccard_bce_loss'
    def __init__(self,weights=[1.,1.],activation='sigmoid'):
        super().__init__()
        self.bce = BCELoss()
        self.jaccard = JaccardLoss(activation=activation)
        self.w1,self.w2=weights[0],weights[1]
    def forward(self,y_pr,y_gt):
        return self.w2*self.bce(y_pr,y_gt) + self.w1*self.jaccard(y_pr, y_gt)
    



class DiceFocalLoss(nn.Module):
    __name__='dice_focal_loss'
    def __init__(self,weights=[1.,1.],activation='sigmoid'):
        super().__init__()
        self.focal = FocalLoss()
        self.dice = DiceLoss(activation=activation)
        self.w1,self.w2=weights[0],weights[1]
    def forward(self,y_pr,y_gt):
        return self.w2*self.focal(y_pr,y_gt) + self.w1*self.dice(y_pr, y_gt)
    


class JaccardFocalLoss(nn.Module):
    __name__='jaccard_focal_loss'
    def __init__(self,weights=[1.,1.],activation='sigmoid'):
        super().__init__()
        self.focal = FocalLoss()
        self.jaccard = JaccardLoss(activation=activation)
        self.w1,self.w2=weights[0],weights[1]
    def forward(self,y_pr,y_gt):
        return self.w2*self.focal(y_pr,y_gt) + self.w1*self.jaccard(y_pr, y_gt)



class DiceFocalBCELoss(nn.Module):
    __name__='dice_focal_bce_loss'
    def __init__(self,weights=[1.,1.,1.],activation='sigmoid'):
        super().__init__()
        self.focal = FocalLoss()
        self.dice = DiceLoss(activation=activation)
        self.bce = BCELoss()
        self.w1,self.w2,self.w3=weights[0],weights[1],weights[2]
    def forward(self,y_pr,y_gt):
        return self.w1*self.dice(y_pr, y_gt) + self.w2*self.focal(y_pr, y_gt) + self.w3*self.bce(y_pr, y_gt)



class JaccardFocalBCELoss(nn.Module):
    __name__='jaccard_focal_bce_loss'
    def __init__(self,weights=[1.,1.,1.],activation='sigmoid'):
        super().__init__()
        self.focal = FocalLoss()
        self.jaccard = JaccardLoss(activation=activation)
        self.bce = BCELoss()
        self.w1,self.w2,self.w3=weights[0],weights[1],weights[2]
    def forward(self,y_pr,y_gt):
        return self.w1*self.jaccard(y_pr, y_gt) + self.w2*self.focal(y_pr, y_gt) + self.w3*self.bce(y_pr, y_gt)




class MultiClass_loss(nn.Module) :
    __name__='multiclass_loss'
    def __init__(self,sloss='Dice',focal=False,entropy=1.,balance = [0.5,0.5],weights = [1.,5.,5.,1.],activation='sigmoid',normalize=True):
        super().__init__()
        #main_loss = []
        self.sloss = sloss
        self.focal = focal
        self.entropy = entropy
        self.act = activation
        self.w1=balance
        self.w2=weights
        self.norm = normalize
        self.loss_list = [self.get() for i in range(len(weights))]
        if(self.entropy>0.):self.ce_loss = nn.BCEWithLogitsLoss(reduction='mean')                                        
    def get(self):
        lname = self.sloss.lower()
        if(self.focal):
            lname += '+focal'
        if(self.entropy):
             lname += '+bce'
        return get_criterion(name=lname,activation = self.act,w =self.w1)[0]
    def forward(self,y_pr,y_gt):
        loss = 0
        assert (len(self.w2) == y_pr.size(1) == y_gt.size(1)),'weights and channel numbers dont match!'
        for i,(wi,lf) in enumerate(list(zip(self.w2,self.loss_list))):
            #loss += wi * lf(y_pr[:,i,...].unsqueeze(1),y_gt[:,i,...].unsqueeze(1))
            loss += wi * lf(y_pr[:,i,...],y_gt[:,i,...])
        #if(self.entropy>0):
           # loss += self.entropy * self.ce_loss(y_pr,y_gt)
        if(self.norm):
                loss /= (np.sum(self.w2))#+self.entropy)
        return loss
        
    

class CombinedLoss(nn.Module):
    def __init__(self,name = 'dice+bce',
                 weights = [0.5,0.5],
                 activation = 'sigmoid',
                 channel_weights = [1.,1.,1.],
                 losses_per_channel = ['dice','bce','jaccard','focal'],
                 normalize = True):
        super().__init__() 
        self.mapping = {
            'dice' : DiceLoss(activation = None),
            'jaccard': JaccardLoss(activation = None),
            'bce'  : BCELoss(),
            'focal' : FocalLoss()
            }
        self.weights = list( weights / np.sum(weights))
        self.activation = Activation(activation)
        self.need_activation = ['bce','focal']
        self.names = name.split('+')
        self.channel_losses = losses_per_channel
        self.channel_weights = channel_weights
        self.__name__ = f'{name.lower()}_loss'
        self.norm = normalize
        
    def forward(self,y_pred,y_target):     
        loss = 0
        num_channels = y_target.size(1)  
        act_y_pred = self.activation(y_pred)     
        for w,name in enumerate(self.names):            
            criterion = self.mapping[name]
            pred_tensor = y_pred if(name in self.need_activation) else act_y_pred          
            if(name in self.channel_losses):            
                for i in range(num_channels):  
                    loss += self.channel_weights[i] * self.weights[w] * criterion(pred_tensor[:,i,...],y_target[:,i,...])
            else:
                loss += self.weights[w] * criterion(pred_tensor,y_target)               
        return loss / sum(self.channel_weights) if(self.norm) else loss
    
    

class DiceLoss2(DiceLoss):

    def __init__(self, eps=1., beta=1., activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return 1 - F_.f_score(
            y_pr, y_gt,
            beta=self.beta,
            eps=self.eps,
            threshold=0.3,
            ignore_channels=self.ignore_channels,
        )

# def get_criterion(name='jaccard',activation='sigmoid',w=[1.,1.,1.]):
#     if(name=='dice'):
#         #from segmentation_models_pytorch.utils.losses import DiceLoss
#         return DiceLoss(activation=activation),1
#     elif(name=='focal'):
#         from Training.losses import FocalLoss
#         return FocalLoss(),1#activation=activation)
#     elif(name == 'jaccard'):
#         from segmentation_models_pytorch.utils.losses import JaccardLoss
#         return JaccardLoss(activation=activation),1
#     elif(name =='bce'):
#         from Training.losses import BCELoss
#         return BCELoss(),1#activation=activation)
#     elif(name=='blovasz'):
#         from Training.lovasz import BinaryLovaszLoss
#         return BinaryLovaszLoss(activation=activation),1
#     elif(name =='lovasz'):
#         from Training.lovasz import LovaszLoss
#         return LovaszLoss(activation=activation),1
#     elif(name == 'dice+bce'):
#         from Training.losses import DiceBCELoss
#         return DiceBCELoss(weights=w[:2]),w[0]+w[1]
#     elif(name == 'jaccard+bce'):
#         from Training.losses import JaccardBCELoss
#         return JaccardBCELoss(weights=w[:2]),w[0]+w[1]
#     elif(name =='dice+focal'):
#         from Training.losses import DiceFocalLoss
#         return DiceFocalLoss(weights=w[:2]),w[0]+w[1]
#     elif(name == 'jaccard+focal'):
#         from Training.losses import JaccardFocalLoss
#         return JaccardFocalLoss(weights=w[:2]),w[0]+w[1]
#     elif(name == 'jaccard+focal+bce'):
#         from Training.losses import JaccardFocalBCELoss
#         return JaccardFocalBCELoss(weights=w),w[0]+w[1]+w[2]
#     elif(name == 'dice+focal+bce'):
#         from Training.losses import DiceFocalBCELoss
#         return DiceFocalBCELoss(weights=w),w[0]+w[1]+w[2]
#     else:
#         raise NotImplementedError()
        
def mixup_data(x, y, alpha=1.0):
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    bs = x.size(0)
    index = torch.randperm(bs)
    mixed_x = lam * x + (1 - lam) * x[index, :]

    return mixed_x, y, y[index], lam

class Mixup_Criterion(nn.Module):
    def __init__(self,criterion:nn.Module):
        super().__init__()
        self.criterion = criterion
        self.__name__ = 'mixup_' + self.criterion.__name__
    def forward(self,y_gt,y_gti,y_pr,lam):
        return lam * self.criterion(y_pr,y_gt) + (1 - lam) * self.criterion(y_pr,y_gti)
def rand_bbox(size, lam):
    W = size[2]
    H = size[3]
    cut_rat = np.sqrt(1. - lam)
    cut_w = np.int(W * cut_rat)
    cut_h = np.int(H * cut_rat)

    # uniform
    cx = np.random.randint(W)
    cy = np.random.randint(H)

    bbx1 = np.clip(cx - cut_w // 2, 0, W)
    bby1 = np.clip(cy - cut_h // 2, 0, H)
    bbx2 = np.clip(cx + cut_w // 2, 0, W)
    bby2 = np.clip(cy + cut_h // 2, 0, H)

    return bbx1, bby1, bbx2, bby2

def cutmix_data(x, y, alpha = 1.0):
    indices = torch.randperm(x.size(0))
    shuffled_y = y[indices]
    lam = np.random.beta(alpha, alpha)
    bbx1, bby1, bbx2, bby2 = rand_bbox(x.size(), lam)
    x[:, :, bbx1:bbx2, bby1:bby2] = x[indices, :, bbx1:bbx2, bby1:bby2]
    target = y
    target[:, :, bbx1:bbx2, bby1:bby2] = target[indices, :, bbx1:bbx2, bby1:bby2]
    # adjust lambda to exactly match pixel ratio
    lam = 1 - ((bbx2 - bbx1) * (bby2 - bby1) / (x.size()[-1] * x.size()[-2]))
    mixed_x = x
    return mixed_x, y , shuffled_y, lam,target
class CutMix_Criterion(nn.Module):
    def __init__(self,criterion:nn.Module):
        super().__init__()
        self.criterion = criterion
        self.__name__ = 'cutmix_' + self.criterion.__name__
    def forward(self,y_gt,y_gti,y_pr,lam):
        return lam * self.criterion(y_pr,y_gt) + (1. - lam) * self.criterion(y_pr,y_gti)
class CutMix_Criterion2(nn.Module):
    def __init__(self,criterion:nn.Module):
        super().__init__()
        self.criterion = criterion
        self.__name__ = 'cutmix_' + self.criterion.__name__
    def forward(self,y_gt_new,y_pr,lam):
        return self.criterion(y_pr,y_gt_new)
'''
class JaccardLoss(nn.Module):
    __name__ = 'jaccard_loss'

    def __init__(self, eps=1e-7, activation='sigmoid'):
        super().__init__()
        self.activation = activation
        self.eps = eps

    def forward(self, y_pr, y_gt):
        return 1 - jaccard(y_pr, y_gt, eps=self.eps, threshold=None, activation=self.activation)


class DiceLoss(nn.Module):
    __name__ = 'dice_loss'

    def __init__(self, eps=1e-7, activation='sigmoid'):
        super().__init__()
        self.activation = activation
        self.eps = eps

    def forward(self, y_pr, y_gt):
        return 1 - f_score(y_pr, y_gt, beta=1., eps=self.eps, threshold=0.5, activation=self.activation)
class Mixed_loss(nn.Module):
    def __init__(self,losses={'jaccard_loss' : {'multiplier' : 2.,'weights':[1.,2.,3.]},
                              'dice_loss' : {'multiplier' : 1.,'weights': [1.]} },reduce = 'sum'):
        super().__init__()
        self.mapping = {'jaccard_loss' : JaccardLoss(),
                        'dice_loss' : DiceLoss(),
                        'focal_loss' : FocalLoss(),
                        'bce_loss' : BCELoss()}
        self.overall_loss = 0
        self.losses = losses
        self.need_sigmoid = {'jaccard_loss','dice_loss','bce_loss'}
        self.reduce = reduce
        self.__name__= 'loss:'
        for l in self.losses:
            k=self.losses[l]['multiplier']
            self.__name__ += f'{k}*{l}_'
    def forward(self,y_pred,y_gt):
        sigmoid_y_pred = torch.sigmoid(y_pred)
        for loss in self.losses:
            this_loss = 0
            channel_losses = []
            loss_name = loss
            loss_func = self.mapping[loss_name]
            weights = self.losses[loss_name]['weights']
            multiplier = self.losses[loss_name]['multiplier']
            channels = y_gt.size(1)
            needs_sigmoid = True if loss_name in self.need_sigmoid else False
            per_channel = True if len(weights)>1 else False
            if(per_channel == True) : 
                assert ( len(channels) == len(weights))
                for i,w in enumerate(weights):
                    channel_losses.append(w*loss_func(sigmoid_y_pred[:,i,...] if needs_sigmoid else y_pred[:,i,...],
                                                      y_gt[:,i,...]))        
                if(self.reduce == 'avg'):
                    this_loss = sum(channel_losses)/channels
                elif(self.reduce == 'sum'):
                    this_loss = sum(channel_losses)
                else:
                    raise NotImplementedError()
            else : this_loss = loss_func(sigmoid_y_pred if needs_sigmoid else y_pred,
                                                      y_gt)
            self.overall_loss += multiplier * this_loss
        return self.overall_loss
'''




# if __name__ == '__main__':

#   #"""
#   x2 = torch.tensor(
#     [
#       [
#         [[0,0.2,0.3],
#           [0.2,0.1,0.1],
#           [0.3,0.5,0.5]]
#       ]
#     ])

#   x = torch.tensor(
#     [
#       [
#         [[0,1.0,1.0],
#         [1.0,1.0,0.0],
#         [1.0,0.0,0.0]]
#       ]
#     ])
  
  
#   mask = (x != 0).bool()
#   crit = MaskedBCELoss()
#   print(x.shape)
#   crit(x2,x,mask)
#   print(x.shape)

#   #pos = x.gt(2)
#   #print(pos.shape)
#   #print(pos)
#   #"""
  



import math
import torch
from torch.optim.optimizer import Optimizer, required
import itertools as it
from .lookahead import *
from .radam import * 

def Ranger(params, alpha=0.5, k=6, *args, **kwargs):
     radam = RAdam(params, *args, **kwargs)
     return Lookahead(radam, alpha, k)





import torch
from torch.nn.functional import interpolate

import os
import cv2
import random

import numpy as np

from math import ceil
from skimage.segmentation import watershed
from skimage.measure import label


def make_dir(path):
    os.makedirs(path,exist_ok=True)

#load model util
def load_model(model,path):
    checkpoint = torch.load(f'{path}/best_model.pth', map_location='cpu')
    model.load_state_dict(checkpoint['state_dict'])
    model.eval()
    return model

#simple send to tensor
def totensor(x):
    return torch.from_numpy(x.transpose(2,0,1)).unsqueeze(0).float()

def normalize(x):
    x = x.astype(np.float64)
    x -= (0.5 * 255.0)
    x /= (0.5 * 255.0)
    return x
    
#padding - unpadding utils
def get_pads(L,div=32):
    if L % div == 0:
        x1,x2 = 0,0
    else:
        L_pad = ceil(L / div) * div
        dL = L_pad - L
        x1 = max(1,dL//2)
        x2 = dL - x1
    return x1,x2

def ratio_resize_pad(img,ratio=None,div=32):

    h_rsz,w_rsz = h_orig,w_orig = img.shape[:2]

    if ratio is not None:
        if ratio != 1.0:
            h_rsz = ceil(h_orig * ratio)
            w_rsz = ceil(w_orig * ratio)
            img = cv2.resize(img,(w_rsz,h_rsz))

    t,b = get_pads(h_rsz,div)
    l,r = get_pads(w_rsz,div)
    img = cv2.copyMakeBorder(img,t,b,l,r,borderType=cv2.BORDER_CONSTANT,value=0.0)

    info = {'orig_size' : (h_orig,w_orig),'pads':(t,b,l,r)}
    return img,info
    
def unpad_resize(img,info):
    h,w = img.shape[2:]

    t,b,l,r = info['pads']
    orig_size = info['orig_size']

    img = img[:,:,t:h-b,l:w-r]
    if h != orig_size[0] or w != orig_size[1]:
        #img = cv2.resize(img,orig_size).astype(np.uint8)
        img = interpolate(img,size=orig_size)
    return img



import unittest

class TestMeters(unittest.TestCase):
    def test_reset(self):
        #testing if the 2 meters are actually resetting
        avg_meter = AverageMeter()
        avg_meter.value = 5.0
        avg_meter.average = 3.0
        avg_meter.count = 10
        avg_meter.reset()
        self.assertEqual(avg_meter.count, 0)
        self.assertEqual(avg_meter.value, 0.0)
        self.assertEqual(avg_meter.average, 0.0)

        short_term_meter = ShortTermMemoryMeter(memory_length=3)
        short_term_meter.value = 5.0
        short_term_meter.length = 3
        short_term_meter.in_memory = [3.0, 4.0, 5.0]
        short_term_meter.average = 4.0
        short_term_meter.reset()
        self.assertEqual(short_term_meter.length, 0)
        self.assertEqual(short_term_meter.value, 0)
        self.assertEqual(short_term_meter.average, 0)
    
    def test_update(self):
        avg_meter = AverageMeter()
        avg_meter.value = 4.0
        avg_meter.count = 1
        avg_meter.average = 4.0
        avg_meter.update(5.0)
        self.assertEqual(avg_meter.count, 2)
        self.assertEqual(avg_meter.value, 5.0)
        self.assertEqual(avg_meter.average, 4.5)

        short_term_meter = ShortTermMemoryMeter(memory_length=3)
        short_term_meter.update(2.0)
        self.assertEqual(short_term_meter.length, 1)
        self.assertEqual(short_term_meter.average, 2.0)
        
class TestSampler(unittest.TestCase):
    def test_reset(self):
        sampler = Sampler(value_names=['value1', 'value2'], rate=2)
        sampler.reset()
        self.assertEqual(sampler.history, {'value1': [], 'value2': []})

    def test_update(self):
        sampler = Sampler(value_names=['value1', 'value2'], rate=2)
        sampler.update([1, 2], 0)
        self.assertEqual(sampler.history, {'value1': [1], 'value2': [2]})
        
class DummyCriterion(torch.nn.Module):
    def __init__(self):
        super(DummyCriterion, self).__init__()
        self.__name__="Module"

    def forward(self, x, y):
        return x + y
    
class DummyMetric:
    def __init__(self):
        self.__name__=""

    def get_update(self):
        return 0.5

class TestMetricsManager(unittest.TestCase):
    def test_initialization(self):
        classes = ['class1', 'class2']
        metric_names = ['iou', 'fscore']
        thresh = 0.5

        manager = MetricsManager(classes, metric_names, thresh)

        self.assertEqual(manager.classes, classes)
        self.assertEqual(manager.metric_names, metric_names)
        self.assertEqual(manager.thresh, thresh)
        self.assertEqual(len(manager.metrics), len(classes) * len(metric_names))

    def test_register_scores(self):
        classes = ['class1', 'class2']
        metric_names = ['iou', 'fscore']
        thresh = 0.5

        manager = MetricsManager(classes, metric_names, thresh)

        pdict = {}
        seg_loss_meter = AverageMeter()
        seg_criterion = DummyCriterion()
        cls_loss_meter = AverageMeter()
        cls_criterion = DummyCriterion()
        cls_accuracy_meter = AverageMeter()
        cls_f1score_meter = AverageMeter()
        dm1 = DummyMetric()
        dm1.__name__="metric1"
        dm2 = DummyMetric()
        dm2.__name__="metric2"
        mnms = [(dm1, AverageMeter()), (dm2, AverageMeter())]

        pdict = manager.register_scores(pdict, seg_loss_meter, seg_criterion, cls_loss_meter, cls_criterion, mnms, cls_accuracy_meter, cls_f1score_meter)

        expected_length = len(classes) * len(metric_names) + 2
        self.assertEqual(len(pdict), expected_length)
        



# from here https://github.com/jxbz/madam/blob/master/pytorch/optim/madam.py
import torch
from torch.optim.optimizer import Optimizer, required


class Madam(Optimizer):

    def __init__(self, params, lr=0.01, p_scale=3.0, g_bound=10.0):

        self.p_scale = p_scale
        self.g_bound = g_bound
        defaults = dict(lr=lr)
        super(Madam, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]
                if len(state) == 0:
                    state['max'] = self.p_scale*(p*p).mean().sqrt().item()
                    state['step'] = 0
                    state['exp_avg_sq'] = torch.zeros_like(p)

                state['step'] += 1
                bias_correction = 1 - 0.999 ** state['step']
                state['exp_avg_sq'] = 0.999 * state['exp_avg_sq'] + 0.001 * p.grad.data**2
                
                g_normed = p.grad.data / (state['exp_avg_sq']/bias_correction).sqrt()
                g_normed[torch.isnan(g_normed)] = 0
                g_normed.clamp_(-self.g_bound, self.g_bound)
                
                p.data *= torch.exp( -group['lr']*g_normed*torch.sign(p.data) )
                p.data.clamp_(-state['max'], state['max'])

        return loss



"""
This code is generated by Ridvan Salih KUZU @DLR
LAST EDITED:  14.09.2021
ABOUT SCRIPT:
It defines a data reader for Sentinel-1 eath observation data
"""

import os
from torch.utils.data import Dataset
import zipfile
import tarfile
from sh import gunzip
from glob import glob
import pickle
import geopandas as gpd
import numpy as np
import rasterio as rio
from rasterio import features
from tqdm import tqdm


class S1Reader(Dataset):
    """
    THIS CLASS INITIALIZES THE DATA READER FOR SENTINEL-1 DATA
    """
    def __init__(self, input_dir, label_dir, label_ids=None, transform=None, min_area_to_ignore = 1000, selected_time_points=None):
        '''
        THIS FUNCTION INITIALIZES DATA READER.
        :param input_dir: directory of input images in zip format
        :param label_dir: directory of ground-truth polygons in GeoJSON format
        :param label_ids: an array of crop IDs in order. if the crop labels in GeoJSON data is not started from index 0 it can be used. Otherwise it is not required.
        :param transform: data transformer function for the augmentation or data processing
        :param min_area_to_ignore: threshold m2 to eliminate small agricultural fields less than a certain threshold. By default, threshold is 1000 m2
        :param selected_time_points: If a sub set of the time series will be exploited, it can determine the index of those times in a given time series dataset

        :return: None
        '''

        self.data_transform = transform
        self.selected_time_points=selected_time_points
        self.crop_ids = label_ids
        if label_ids is not None and not isinstance(label_ids, list):
            self.crop_ids = label_ids.tolist()

        self.npyfolder = input_dir.replace(".zip", "/time_series")
        #self.npyfolder = "/mnt/c/users/ali/documents/cnrs-l internship 2021/ai4eo data/ref_fusion_competition_south_africa_train_source_sentinel_1/ref_fusion_competition_south_africa_train_source_sentinel_1_34S_19E_258N_asc_34S_19E_258N_2017/"
        self.labels = S1Reader._setup(input_dir, label_dir,self.npyfolder,min_area_to_ignore)

    def __len__(self):
        """
        THIS FUNCTION RETURNS THE LENGTH OF DATASET
        """
        return len(self.labels)

    def __getitem__(self, item):
        """
        THIS FUNCTION ITERATE OVER THE DATASET BY GIVEN ITEM NO AND RETURNS FOLLOWINGS:
        :return: image_stack in size of [Time Stamp, Image Dimension (Channel), Height, Width] , crop_label, field_mask in size of [Height, Width], field_id
        """

        feature = self.labels.iloc[item]

        npyfile = os.path.join(self.npyfolder, "fid_{}.npz".format(feature.fid))
        if os.path.exists(npyfile): # use saved numpy array if already created
            try:
                object = np.load(npyfile)
                image_stack = object["image_stack"]
                mask = object["mask"]
            except zipfile.BadZipFile:
                print("ERROR: {} is a bad zipfile...".format(npyfile))
                raise
        else:
            print("ERROR: {} is a missing...".format(npyfile))
            raise

        if self.data_transform is not None:
            image_stack, mask = self.data_transform(image_stack, mask)

        if self.selected_time_points is not None:
            image_stack = image_stack[self.selected_time_points]

        if self.crop_ids is not None:
            label = self.crop_ids.index(feature.crop_id)
        else:
            label = feature.crop_id

        return image_stack, label, mask, feature.fid


    @staticmethod
    def _setup(rootpath, labelgeojson, npyfolder, min_area_to_ignore=1000):
        """
        THIS FUNCTION PREPARES THE PLANET READER BY SPLITTING AND RASTERIZING EACH CROP FIELD AND SAVING INTO SEPERATE FILES FOR SPEED UP THE FURTHER USE OF DATA.

        This utility function unzipps a dataset and performs a field-wise aggregation.
        results are written to a .npz cache with same name as zippath

        :param rootpath: directory of input images
        :param labelgeojson: directory of ground-truth polygons in GeoJSON format
        :param npyfolder: folder to save the field data for each field polygon
        :param min_area_to_ignore: threshold m2 to eliminate small agricultural fields less than a certain threshold. By default, threshold is 1000 m2
        :return: labels of the saved fields
        """

        with open(os.path.join(rootpath, "bbox.pkl"), 'rb') as f:
            bbox = pickle.load(f)
            crs = str(bbox.crs)
            minx, miny, maxx, maxy = bbox.min_x, bbox.min_y, bbox.max_x, bbox.max_y

        labels = gpd.read_file(labelgeojson)
        TO_DROP=[80172,80217,80225,80228,80593,80602,80609,80610,81060,81062,81065,81123,81148,81149,81150,
        81178,81242,81605,81606,81616,185234,185256,185261,185262,185279,185281,185282,185652,185653,185682,185750,185916,78030,83416,83451,83485,84068,80985]
        labels = labels[~labels.fid.isin(TO_DROP)]
        # project to same coordinate reference system (crs) as the imagery
        ignore = labels.geometry.area > min_area_to_ignore
        print(f"INFO: Ignoring {(~ignore).sum()}/{len(ignore)} fields with area < {min_area_to_ignore}m2")
        labels = labels.loc[ignore]
        labels.reset_index(drop = True, inplace=True)
        labels = labels.to_crs(crs) #TODO: CHECK IF NECESSARY

        vv = np.load(os.path.join(rootpath, "vv.npy"))
        vh = np.load(os.path.join(rootpath, "vh.npy"))
        ia = np.load(os.path.join(rootpath, "angle.npy"))
        bands = np.stack([vv[:,:,:,0],vh[:,:,:,0],ia[:,:,:,0]], axis=3)
        _, width, height, _ = bands.shape

        bands=bands.transpose(0, 3, 1, 2)

        transform = rio.transform.from_bounds(minx, miny, maxx, maxy, width, height)

        fid_mask = features.rasterize(zip(labels.geometry, labels.fid), all_touched=True,
                                  transform=transform, out_shape=(width, height))
        assert len(np.unique(fid_mask)) > 0, f"WARNING: Vectorized fid mask contains no fields. " \
                                             f"Does the label geojson {labelgeojson} cover the region defined by {rootpath}?"

        crop_mask = features.rasterize(zip(labels.geometry, labels.crop_id), all_touched=True,
                                  transform=transform, out_shape=(width, height))
        assert len(np.unique(crop_mask)) > 0, f"WARNING: Vectorized fid mask contains no fields. " \
                                              f"Does the label geojson {labelgeojson} cover the region defined by {rootpath}?"


        for index, feature in tqdm(labels.iterrows(), total=len(labels), position=0, leave=True, desc="INFO: Extracting time series into the folder: {}".format(npyfolder)):

            npyfile = os.path.join(npyfolder, "fid_{}.npz".format(feature.fid))
            if not os.path.exists(npyfile):

                left, bottom, right, top = feature.geometry.bounds
                window = rio.windows.from_bounds(left, bottom, right, top, transform)

                row_start = round(window.row_off)
                row_end = round(window.row_off) + round(window.height)
                col_start = round(window.col_off)
                col_end = round(window.col_off) + round(window.width)

                image_stack = bands[:, :,row_start:row_end, col_start:col_end]
                mask = fid_mask[row_start:row_end, col_start:col_end]
                mask[mask != feature.fid] = 0
                mask[mask == feature.fid] = 1
                os.makedirs(npyfolder, exist_ok=True)
                np.savez(npyfile, image_stack=image_stack.astype(np.float32), mask=mask.astype(np.float32), feature=feature.drop("geometry").to_dict())

        return labels




if __name__ == '__main__':
    """
    EXAMPLE USAGE OF DATA READER
    """

    rootpath = "../data/dlr_fusion_competition_germany_train_source_sentinel_1/dlr_fusion_competition_germany_train_source_sentinel_1_33N_18E_242N_2018/"
    labelgeojson = "../data/dlr_fusion_competition_germany_train_labels/dlr_fusion_competition_germany_train_labels_33N_18E_242N/labels.geojson"
    ds = S1Reader(rootpath, labelgeojson,selected_time_points=[1,2,3])
    X,y,m,fid = ds[0]



import unittest
import torch
from torch import nn
from . import get_optimizer,optimizer_mapping
import numpy as np


class TestOptimizers(unittest.TestCase):
    def test_getoptimzer_return(self):
        for optimizer in optimizer_mapping:
            net=nn.Sequential(
                nn.Linear(2,1),
            )
            optim=get_optimizer(optimizer,net.parameters(),lr=1e-3)
            optim.zero_grad()
            optim.step()
            self.assertIsInstance(optim,torch.optim.Optimizer,f"{optimizer} did not return an torch.optim.Optimizer")

    def test_getoptimzer_lookAhead(self):
        for optimizer in optimizer_mapping:
            net=nn.Sequential(
                nn.Linear(2,1),
            )
            if optimizer =="adan": #TODO fix adan
                with self.assertRaises(AssertionError) as context:
                    optim=get_optimizer(optimizer,net.parameters(),lr=1e-3,lookAhead=True)
            else:       
                optim=get_optimizer(optimizer,net.parameters(),lr=1e-3,lookAhead=True)
                optim.zero_grad()
                optim.step()

                self.assertIsInstance(optim,torch.optim.Optimizer,f"{optimizer} did not return an torch.optim.Optimizer")

    def test_keys_lower_case(self):
        for optim in optimizer_mapping:
            self.assertEqual(optim,optim.lower())

    def test_getoptimzer_functionality(self):
        for optimizer in optimizer_mapping:
            for __ in range(5):
                net=nn.Sequential(
                nn.Linear(2,1),
                )
                optim=get_optimizer(optimizer,net.parameters(),lr=1e-3)
                loss1=[]
                for _ in range(500):
                    optim.zero_grad()
                    X=torch.randint(0,10,(50,2),dtype=torch.float)
                    y=X.sum(axis=1)
                    loss=nn.MSELoss()(y,net(X))
                    loss.backward()
                    loss1.append(loss.item())
                    optim.step()
                loss2=[]
                for _ in range(500):
                    optim.zero_grad()
                    X=torch.randint(0,10,(50,2),dtype=torch.float)
                    y=X.sum(axis=1)
                    loss=nn.MSELoss()(y,net(X))
                    loss.backward()
                    loss2.append(loss.item())
                    optim.step()
                self.assertGreaterEqual(np.array(loss1).mean(),np.array(loss2).mean(),f"{optimizer} is not functioning as an optimizer")



if __name__=="__main__":
    unittest.main()



"""
This code is generated by Ridvan Salih KUZU @DLR
LAST EDITED:  14.09.2021
ABOUT SCRIPT:
It defines Data Loader for training, validation and test partitions using torch and numpy
"""

import numpy as np
import torch
from sklearn.model_selection import train_test_split
import geopandas as gpd

class DataLoader():
    """
    THIS CLASS INITIALIZES THE TRAINING, VALIDATION, AND TEST DATA LOADERS
    """
    def __init__(self, train_val_reader=None, test_reader=None, validation_split=0.2):
        '''
        THIS FUNCTION INITIALIZES THE DATA LOADER.
        :param train_val_reader: data reader inherited from torch.utils.data.Dataset for train/validation partitions
        :param test_reader: data reader inherited from torch.utils.data.Dataset for test partition
        :param validation_split: the rate of validation data (between 0.0-1.0) to split data into 2 partitions for training and validation
        :return: None
        '''

        if train_val_reader is not None:
            # indices = list(range(len(train_val_reader)))
            # np.random.RandomState(0).shuffle(indices)
            # split = int(np.floor(validation_split * len(train_val_reader)))
            # train_indices, val_indices = indices[split:], indices[:split]
            #By ali
            combined_labels_dir = "/mnt/c/users/ali/documents/cnrs-l internship 2021/ai4eo data/ref_fusion_competition_south_africa_train_labels/combined/labels.geojson"
            TO_DROP=[80172,80217,80225,80228,80593,80602,80609,80610,81060,81062,81065,81123,81148,81149,81150,
            81178,81242,81605,81606,81616,185234,185256,185261,185262,185279,185281,185282,185652,185653,185682,185750,185916,78030,83416,83451,83485,84068,80985]
            labels = gpd.read_file(combined_labels_dir)
            labels = labels[~labels.fid.isin(TO_DROP)] 
            labels.reset_index(drop = True, inplace=True)
            x = labels['fid'].tolist()
            y = labels['crop_id'].tolist()
            X_train, X_valid, y_train, y_valid = train_test_split(
                    x, y,stratify=y, test_size=validation_split)
            train_indices = labels.loc[labels['fid'].isin(X_train)].index.tolist()
            valid_indices = labels.loc[labels['fid'].isin(X_valid)].index.tolist()
            # Splitting the data into training and validation partitions
            self.train_dataset = torch.utils.data.Subset(train_val_reader, train_indices)
            self.val_dataset = torch.utils.data.Subset(train_val_reader, valid_indices)

        if test_reader is not None:
            self.test_dataset=test_reader

        if test_reader is None and train_val_reader is None:
            raise

    def get_train_loader(self, batch_size=4, num_workers=2):
        '''
        THIS FUNCTION RETURNS THE TRAINING DATA LOADER.
        :param batch_size: number of batches while loading the training data
        :param num_workers: number of workers to operating in parallel
        :return: torch.utils.data.DataLoader
        '''
        print('INFO: Training data loader initialized.')
        return  torch.utils.data.DataLoader(self.train_dataset, batch_size=batch_size, num_workers=num_workers)


    def get_validation_loader(self, batch_size, num_workers):
        '''
        THIS FUNCTION RETURNS THE VALIDATION DATA LOADER.
        :param batch_size: number of batches while loading the validation data
        :param num_workers: number of workers to operating in parallel
        :return: torch.utils.data.DataLoader
        '''
        print('INFO: Validation data loader initialized.')
        return  torch.utils.data.DataLoader(self.val_dataset, batch_size=batch_size, num_workers=num_workers)

    def get_test_loader(self, batch_size, num_workers):
        '''
        THIS FUNCTION RETURNS THE TEST DATA LOADER.
        :param batch_size: number of batches while loading the test data
        :param num_workers: number of workers to operating in parallel
        :return: torch.utils.data.DataLoader
        '''
        print('INFO: Test data loader initialized.')
        return  torch.utils.data.DataLoader(self.test_dataset, batch_size=batch_size, num_workers=num_workers)



# https://raw.githubusercontent.com/sail-sg/Adan/main/adan.py

# Copyright 2022 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import torch
from torch.optim.optimizer import Optimizer
from timm.utils import *


class Adan(Optimizer):
    """
    Implements a pytorch variant of Adan

    Adan was proposed in
    Adan: Adaptive Nesterov Momentum Algorithm for Faster Optimizing Deep Models[J]. arXiv preprint arXiv:2208.06677, 2022.
    https://arxiv.org/abs/2208.06677
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining parameter groups.
        lr (float, optional): learning rate. (default: 1e-3)
        betas (Tuple[float, float, flot], optional): coefficients used for computing 
            running averages of gradient and its norm. (default: (0.98, 0.92, 0.99))
        eps (float, optional): term added to the denominator to improve 
            numerical stability. (default: 1e-8)
        weight_decay (float, optional): decoupled weight decay (L2 penalty) (default: 0)
        max_grad_norm (float, optional): value used to clip 
            global grad norm (default: 0.0 no clip)
        no_prox (bool): how to perform the decoupled weight decay (default: False)
    """

    def __init__(self, params, lr=1e-3, betas=(0.98, 0.92, 0.99), eps=1e-8,
                 weight_decay=0.0, max_grad_norm=0.0, no_prox=False):
        if not 0.0 <= max_grad_norm:
            raise ValueError("Invalid Max grad norm: {}".format(max_grad_norm))
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= betas[2] < 1.0:
            raise ValueError("Invalid beta parameter at index 2: {}".format(betas[2]))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay,
                        max_grad_norm=max_grad_norm, no_prox=no_prox)
        super(Adan, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(Adan, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('no_prox', False)

    @torch.no_grad()
    def restart_opt(self):
        for group in self.param_groups:
            group['step'] = 0
            for p in group['params']:
                if p.requires_grad:
                    state = self.state[p]
                    # State initialization

                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p)
                    # Exponential moving average of gradient difference
                    state['exp_avg_diff'] = torch.zeros_like(p)

    @torch.no_grad()
    def step(self):
        """
            Performs a single optimization step.
        """
        if self.defaults['max_grad_norm'] > 0:
            device = self.param_groups[0]['params'][0].device
            global_grad_norm = torch.zeros(1, device=device)

            max_grad_norm = torch.tensor(self.defaults['max_grad_norm'], device=device)
            for group in self.param_groups:

                for p in group['params']:
                    if p.grad is not None:
                        grad = p.grad
                        global_grad_norm.add_(grad.pow(2).sum())

            global_grad_norm = torch.sqrt(global_grad_norm)

            clip_global_grad_norm = torch.clamp(max_grad_norm / (global_grad_norm + group['eps']), max=1.0)
        else:
            clip_global_grad_norm = 1.0

        for group in self.param_groups:
            beta1, beta2, beta3 = group['betas']
            # assume same step across group now to simplify things
            # per parameter step can be easily support by making it tensor, or pass list into kernel
            if 'step' in group:
                group['step'] += 1
            else:
                group['step'] = 1

            bias_correction1 = 1.0 - beta1 ** group['step']

            bias_correction2 = 1.0 - beta2 ** group['step']

            bias_correction3 = 1.0 - beta3 ** group['step']

            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]
                if len(state) == 0:
                    state['exp_avg'] = torch.zeros_like(p)
                    state['exp_avg_sq'] = torch.zeros_like(p)
                    state['exp_avg_diff'] = torch.zeros_like(p)

                grad = p.grad.mul_(clip_global_grad_norm)
                if 'pre_grad' not in state or group['step'] == 1:
                    state['pre_grad'] = grad

                copy_grad = grad.clone()

                exp_avg, exp_avg_sq, exp_avg_diff = state['exp_avg'], state['exp_avg_sq'], state['exp_avg_diff']
                diff = grad - state['pre_grad']

                update = grad + beta2 * diff
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)  # m_t
                exp_avg_diff.mul_(beta2).add_(diff, alpha=1 - beta2)  # diff_t
                exp_avg_sq.mul_(beta3).addcmul_(update, update, value=1 - beta3)  # n_t

                denom = ((exp_avg_sq).sqrt() / math.sqrt(bias_correction3)).add_(group['eps'])
                update = ((exp_avg / bias_correction1 + beta2 * exp_avg_diff / bias_correction2)).div_(denom)

                if group['no_prox']:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                    p.add_(update, alpha=-group['lr'])
                else:
                    p.add_(update, alpha=-group['lr'])
                    p.data.div_(1 + group['lr'] * group['weight_decay'])

                state['pre_grad'] = copy_grad



# Standard libraries
import matplotlib.pyplot as plt

import numpy as np
import os
from time import time

from sklearn import metrics
from pathlib import Path
import glob
import tqdm
import geopandas as gpd
import torch
import gc
import cv2
from shapely.geometry import shape


from ToBeChecked.utility.utils import ratio_resize_pad
import rasterio
from rasterio.features import shapes as rio_shapes
from rasterio.enums import ColorInterp
from rasterio.plot import show
from rasterio.windows import Window

# Define some constants


# Fix random seed for reproducibility
np.random.seed(42)


class LargeTiffLoader:
    """
    The primary purpose of this class is to load and present a sizable Tiff image as a 
    collection of smaller, consistently sized images. The `pre_load` method is specifically 
    designed to carry out this operation. Additionally, the `load_index` function facilitates
    the creation and batching of images from a larger image by specifying the pixel positions to be cropped.
    """
   
    def __init__(self,image_directory,image_suffix='.tif'):
        """
        @param image_directory = path of directory containing images to be loaded
        @param image_suffix = type of images entered, default='.tif'
        """
        
        self.image_directory=image_directory
        self.image_suffix=image_suffix


        
    def load_index(self,save_path,col_off, row_off, width, height):

        """
        @param save_path = path of the directory to save cropped image
        @param col_off = indicates the starting column position from which the batch image should begin.
        @param row_off = specifies the starting row position from which the batch image should start.
        @param width = width of the batch image
        @param height = height of the batch image
        """
        for file in os.listdir(self.image_directory):
             if file.endswith(self.image_suffix):
                name = file.split('.')[0]
                with rasterio.open(f"{self.image_directory}/{file}") as src:
                    original_profile = src.profile
                    window = Window(col_off, row_off, width, height)
                    # fragment = src.read(1, window=window)
                    
                    source_colorinterp = dict(zip(src.colorinterp, src.indexes))

                    rgb_indexes = [
                        source_colorinterp[ci]
                        for ci in (ColorInterp.red, ColorInterp.green, ColorInterp.blue)
                    ]
                    data = src.read(rgb_indexes, window=window)

                    show(data, transform=src.window_transform(window))
                   
                    fragment_profile = original_profile.copy()
                    fragment_profile['width'] = data.shape[1]
                    fragment_profile['height'] = data.shape[2]
             
                    fragment_profile['transform'] = src.window_transform(window)
                 
                
                    fragment_output_path = os.path.join(save_path, f'{name}.tif')
                    with rasterio.open(fragment_output_path, 'w',**fragment_profile) as dst:
                        dst.write(data)



    def pre_load(self,mask_directory,mask_suffix='.tif', fragment_size=1024, PATCH_SIZE=1024, STRIDE_SIZE=512, DOWN_SAMPLING=1,transform=None):
        """
        @param mask_directory = path of directory contains masks
        @param mask_suffix = type of mask file, default='.tif'
        @param fragment_size = size of the copped batch image
        @param PATCH_SIZE = patch size 
        @param STRIDE_SIZE = stride size 
        @param DOWN_SAMPLING = downsampling factor
        @param transform = tansformation information of the images or the masks

        Return:
        loaded_images = list of loaded images as numpy array each
        loaded_masks = list of loaded masks as numpy array each
        """
        
        loaded_images=[]
        loaded_masks=[]
        for file in os.listdir(self.image_directory):
            filename = os.fsdecode(file)
            if filename.endswith(self.image_suffix):
                name=filename.split('.')[0]
                
                raster_file = rasterio.open(f'{self.image_directory}/{filename}')
               
                full_img = raster_file.read([1,2,3]).transpose(1,2,0)
                
                HEIGHT_orig, WIDTH_orig = full_img.shape[:2]
                # if self.mask_suffix==".shp":    
                #     with rio.open(f"{self.image_directory}/{filename}") as src:
                #         transform=src.transform

                #     mask_shp = gpd.read_file(f'{self.mask_directory}/{name}{self.mask_suffix}')
                    
                #     mask=poly_conv.convert_polygon_to_mask(mask_shp['geometry'],(HEIGHT_orig,WIDTH_orig),transform=transform)
                   
                # elif self.mask_suffix==".tiff":
                #     mask = rio.open(glob.glob(f'{self.mask_directory}/{name}{self.mask_suffix}'))
                #     mask = mask.read()[0]#.transpose(1,2,0)
                # else:
                #     print("provide .tiff or .shp mask file")  
                #     return
              
                mask = rasterio.open(glob.glob(f'{mask_directory}/{name}{mask_suffix}')[0])
                mask = mask.read()[0]#.transpose(1,2,0)
                

                full_img = cv2.resize(full_img, (WIDTH_orig//DOWN_SAMPLING, HEIGHT_orig//DOWN_SAMPLING))

                #Use below for gray images only
                #full_img = raster_file.read().transpose(1,2,0)[:,:,0]
                #full_img = cv2.cvtColor(full_img,cv2.COLOR_GRAY2RGB)

                full_img, rrp_info = ratio_resize_pad(full_img, ratio = None, div=fragment_size)
                full_mask, mask_rrp_info = ratio_resize_pad(mask, ratio = None, div=fragment_size)


                HEIGHT, WIDTH = full_img.shape[:2]


                full_mask = np.zeros((HEIGHT, WIDTH), dtype=np.float32)
                full_mask[...] = np.nan

                a = 0
                M = 0
                patched_image=[]
                patched_mask=[]
                for hs in range(a,HEIGHT,STRIDE_SIZE):

                    for ws in range(a,WIDTH,STRIDE_SIZE):

                        he = hs+PATCH_SIZE
                        we = ws+PATCH_SIZE
                        patch = full_img[hs:he,ws:we,:]
                        patch_mask = full_mask[hs:he,ws:we]
                      
                        shapes = rio_shapes(patch_mask)
                        geometry = []
                        for shapedict, value in shapes:
                            if value == 0:
                                continue
                            geometry.append(shape(shapedict))

                        patch_gdf = gpd.GeoDataFrame({'geometry': geometry})
                       
                        if len(patch_gdf) == 0:
                            full_mask[hs:he,ws:we] = 0
                        else:
                            patched_image.append(patch)
                            patched_mask.append(patch_gdf)


                            # y_pred = y_pred.detach().cpu().long().numpy()[:,0,:,:].astype(np.int16)

                            # n_patch,_,_ = y_pred.shape
                            # b_ids = np.arange(n_patch) + 1
                            # b_ids = b_ids[:,np.newaxis,np.newaxis]

                            # y_pred_mask = (y_pred.copy().sum(axis=0) > 0).astype(np.int16)
                            # y_pred *= b_ids
                            # y_pred = y_pred.max(axis=0) + M*y_pred_mask
                            # M = y_pred.max()
                            # full_mask[hs:he,ws:we] = y_pred

              
                loaded_images.append(patched_image)
                loaded_masks.append(patched_mask)
                
                return loaded_images,loaded_masks


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 13:25:24 2021

@author: hasan
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

""" Atruos Spatal Pyramid Pooling Module"""

class ASPPConv(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel, dilation = 1):
        modules = [
            nn.Conv2d(in_channels, 
                      out_channels,
                      kernel,
                      padding=dilation if(kernel>1) else 0,
                      dilation=dilation,
                      bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        ]
        super(ASPPConv, self).__init__(*modules)
        
class SeparableASPPConv(nn.Sequential):
    def __init__(self,in_channels,out_channels,kernel,dilation = 1):
        depthwise_conv = nn.Conv2d(in_channels,
                                   in_channels,
                                   kernel,
                                   padding = dilation,
                                   dilation = dilation,
                                   groups = in_channels,
                                   bias = False
                                   )
        pointwise_conv  = nn.Conv2d(in_channels,
                                    out_channels,
                                    kernel_size = 1,
                                    bias = False)
        super().__init__(depthwise_conv,
                         pointwise_conv,
                         nn.BatchNorm2d(out_channels),
                         nn.ReLU(inplace=True))
        
class ASPPPooling(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(ASPPPooling, self).__init__()
        self.module = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
            )

    def forward(self, x):
        batch_size,C,H,W = x.shape
        x = self.module(x)
        #print(x.shape)
        x = F.interpolate(x, size=(H,W), mode='bilinear', align_corners=True)
        return x


class ASPP(nn.Module):
    def __init__(self, in_channels,
                 out_channels=256,
                 atrous_rates = [6,12,18],
                 dropout_rate = 0,
                 separable = False):
        super(ASPP, self).__init__()
        
        modules = []
        modules.append(ASPPConv(in_channels, out_channels, 1))
        _c = SeparableASPPConv if(separable) else ASPPConv
        for rate in atrous_rates:
            
            modules.append(_c(in_channels, out_channels, 3,  rate))

        modules.append(ASPPPooling(in_channels, out_channels))

        self.convs = nn.ModuleList(modules)

        self.project = nn.Sequential(ASPPConv(len(self.convs) * out_channels, out_channels, 1),
                                     nn.Dropout(dropout_rate))

    def forward(self, x):
        res = []
        for i,conv in enumerate(self.convs):
            #print(i)
            res.append(conv(x))
            #print(res[i].shape)
        res = torch.cat(res, dim=1)
        return self.project(res)


class DenseASPPConv(nn.Module):
    def __init__(self,in_channels,
                 mid_channels,
                 out_channels,
                 rate = 3, 
                 separable = False,
                 dropout_rate = 0.0):
        super().__init__()
        self.conv1x1 = ASPPConv(in_channels = in_channels, out_channels = mid_channels, kernel = 1)
        _c = SeparableASPPConv if(separable) else ASPPConv
        self.dilated_conv = _c(in_channels = mid_channels,
                               out_channels = out_channels,
                               kernel = 3,
                               dilation = rate)
        self.drop = nn.Dropout(dropout_rate)
    def forward(self,x):
        x = self.conv1x1(x)
        x = self.dilated_conv(x)
        return self.drop(x)
    
class DenseASPP(nn.Module):
    def __init__(self,
                 in_channels,
                 mid_channels = 256,
                 inter_channels=256,
                 atrous_rates = [3,6,12,18],
                 dropout_rate = 0,
                 separable = False):
        super().__init__()
        modules = []
        for i,rate in enumerate(atrous_rates):
            modules.append(
                            DenseASPPConv(in_channels = in_channels + i * inter_channels,
                                          mid_channels = mid_channels,
                                          out_channels = inter_channels,
                                          rate = rate,
                                          separable = separable,
                                          dropout_rate=dropout_rate
                                )
                            )
        self.out_channels = in_channels + len(atrous_rates) * inter_channels
        self.convs = nn.ModuleList(modules)
    def forward(self,feats):
        for conv in self.convs:
            x = conv(feats)
            feats = torch.cat([feats,x],dim = 1)
        return feats
        
    
'''    
x = torch.rand((2,128,32,32)).cuda()
m = DenseASPP(128,32,64).cuda()
<<<<<<< HEAD
print(m.out_channels)
y = m(x)
print(y.size())


y = m(x)
'''

